"use strict";
var __assign = (this && this.__assign) || Object.assign || function(t) {
    for (var s, i = 1, n = arguments.length; i < n; i++) {
        s = arguments[i];
        for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
            t[p] = s[p];
    }
    return t;
};
Object.defineProperty(exports, "__esModule", { value: true });
var utils_1 = require("./utils");
var matchesState_1 = require("./matchesState");
exports.matchesState = matchesState_1.default;
var mapState_1 = require("./mapState");
exports.mapState = mapState_1.default;
var State_1 = require("./State");
exports.State = State_1.default;
var STATE_DELIMITER = ".";
var HISTORY_KEY = "$history";
var StateNode = /** @class */ (function () {
    function StateNode(config) {
        var _this = this;
        this._relativeValue = new Map();
        this.key = config.key || "(machine)";
        this.parent = config.parent;
        this.id = this.parent
            ? this.parent.id + STATE_DELIMITER + this.key
            : this.key;
        this.initial = config.initial;
        this.parallel = !!config.parallel;
        this.states = config.states
            ? utils_1.mapValues(config.states, function (stateConfig, key) {
                return new StateNode(__assign({}, stateConfig, { key: key, parent: _this }));
            })
            : {};
        this.on = config.on;
    }
    StateNode.prototype.transition = function (state, action) {
        var nextStateValue = this.transitionStateValue(state, action);
        if (!nextStateValue) {
            return undefined;
        }
        return new State_1.default(nextStateValue, State_1.default.from(state));
    };
    StateNode.prototype.transitionStateValue = function (state, action) {
        var _this = this;
        var history = State_1.default.from(state).history;
        var stateValue = utils_1.toTrie(state instanceof State_1.default ? state.value : state);
        if (typeof stateValue === "string") {
            if (!this.states[stateValue]) {
                throw new Error("state doesnt exist");
            }
            var subState = this.states[stateValue];
            var initialState = subState.initialState;
            if (initialState) {
                stateValue = (_a = {}, _a[stateValue] = initialState, _a);
            }
            else {
                return (subState.next(action, history ? history.value : undefined) ||
                    undefined);
            }
        }
        var nextStateValue = utils_1.mapValues(stateValue, function (subStateValue, subStateKey) {
            if (!_this.states[subStateKey]) {
                return undefined;
            }
            var subHistory = history ? history.value[subStateKey] : undefined;
            var subState = new State_1.default(subStateValue, subHistory ? State_1.default.from(subHistory) : undefined);
            return _this.states[subStateKey].transitionStateValue(subState, action);
        });
        if (Array.prototype.every.call(Object.keys(nextStateValue), function (key) {
            return nextStateValue[key] === undefined;
        })) {
            if (this.parallel) {
                return undefined;
            }
            var subStateKey = Object.keys(nextStateValue)[0];
            return this.states[subStateKey].next(action, history ? history.value : undefined);
        }
        if (this.parallel) {
            nextStateValue = __assign({}, this.initialState, nextStateValue);
        }
        return utils_1.mapValues(nextStateValue, function (value, key) {
            if (value) {
                return value;
            }
            return stateValue[key];
        });
        var _a;
    };
    StateNode.prototype.next = function (action, history) {
        var actionType = utils_1.getActionType(action);
        if (!this.on || !this.on[actionType]) {
            return undefined;
        }
        var nextPath = utils_1.toStatePath(this.on[actionType]);
        var currentState = this.parent;
        var currentHistory = history;
        var currentPath = this.key;
        nextPath.forEach(function (subPath) {
            if (!currentState || !currentState.states) {
                throw new Error("Unable to read '" + subPath + "'");
            }
            if (subPath === HISTORY_KEY) {
                if (currentHistory) {
                    subPath =
                        typeof currentHistory === "object"
                            ? Object.keys(currentHistory)[0]
                            : currentHistory;
                }
                else if (currentState.initial) {
                    subPath = currentState.initial;
                }
                else {
                    throw new Error("Cannot read '" + HISTORY_KEY + "' from state '" + currentState.id + "': missing 'initial'");
                }
            }
            if (typeof subPath === "object") {
                subPath = Object.keys(subPath)[0];
            }
            currentState = currentState.states[subPath];
            if (currentState === undefined) {
                throw new Error("Action '" + action + "' on state '" + currentPath + "' leads to undefined state '" + nextPath + "'.");
            }
            currentPath = subPath;
            if (currentHistory) {
                currentHistory = currentHistory[subPath];
            }
        });
        if (!currentState) {
            throw new Error("no state");
        }
        while (currentState.initial) {
            if (!currentState || !currentState.states) {
                throw new Error("Invalid initial state");
            }
            currentState = currentState.states[currentState.initial];
        }
        return currentState.getRelativeValue(this.parent);
    };
    StateNode.prototype.getInitialState = function () {
        console.warn("machine.getInitialState() will be deprecated in 2.0. Please use machine.initialState instead.");
        return this.initialState;
    };
    Object.defineProperty(StateNode.prototype, "initialState", {
        get: function () {
            this._initialState =
                this._initialState ||
                    (this.parallel
                        ? utils_1.mapValues(this.states, function (state) { return state.initialState; })
                        : this.initial);
            return this._initialState;
        },
        enumerable: true,
        configurable: true
    });
    StateNode.prototype.getState = function (relativeStateId) {
        var statePath = utils_1.toStatePath(relativeStateId);
        try {
            return statePath.reduce(function (subState, subPath) {
                if (!subState.states) {
                    throw new Error("Cannot retrieve subPath '" + subPath + "' from node with no states");
                }
                return subState.states[subPath];
            }, this);
        }
        catch (e) {
            return undefined;
        }
    };
    Object.defineProperty(StateNode.prototype, "events", {
        get: function () {
            if (this._events) {
                return this._events;
            }
            var states = this.states;
            var events = new Set(this.on ? Object.keys(this.on) : undefined);
            if (states) {
                Object.keys(states).forEach(function (stateId) {
                    var state = states[stateId];
                    if (state.states) {
                        for (var _i = 0, _a = state.events; _i < _a.length; _i++) {
                            var event_1 = _a[_i];
                            events.add(event_1);
                        }
                    }
                });
            }
            return (this._events = Array.from(events));
        },
        enumerable: true,
        configurable: true
    });
    StateNode.prototype.getRelativeValue = function (toNode) {
        var memoizedRelativeValue = toNode
            ? this._relativeValue.get(toNode)
            : undefined;
        if (memoizedRelativeValue) {
            return memoizedRelativeValue;
        }
        var initialState = this.initialState;
        var relativeValue = initialState
            ? (_a = {},
                _a[this.key] = initialState,
                _a) : this.key;
        var currentNode = this.parent;
        while (currentNode && currentNode !== toNode) {
            var currentInitialState = currentNode.initialState;
            relativeValue = (_b = {},
                _b[currentNode.key] = typeof currentInitialState === "object" &&
                    typeof relativeValue === "object"
                    ? __assign({}, currentInitialState, relativeValue) : relativeValue,
                _b);
            currentNode = currentNode.parent;
        }
        this._relativeValue.set(toNode, relativeValue);
        return relativeValue;
        var _a, _b;
    };
    return StateNode;
}());
exports.StateNode = StateNode;
function Machine(config) {
    return new StateNode(config);
}
exports.Machine = Machine;
